# RCAIDE/Methods/Noise/Correlation_Buildup/Turbofan/turbofan_engine_noise.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

 # RCAIDE imports 
from RCAIDE.Framework.Core                import Units , Data  
from .angle_of_attack_effect    import angle_of_attack_effect
from .external_plug_effect      import external_plug_effect
from .ground_proximity_effect   import ground_proximity_effect
from .jet_installation_effect   import jet_installation_effect
from .mixed_noise_component     import mixed_noise_component 
from .primary_noise_component   import primary_noise_component
from .secondary_noise_component import secondary_noise_component

from RCAIDE.Library.Methods.Noise.Common  import noise_tone_correction
from RCAIDE.Library.Methods.Noise.Common  import atmospheric_attenuation 
from RCAIDE.Library.Methods.Noise.Common  import SPL_arithmetic
from RCAIDE.Library.Methods.Noise.Metrics import PNL_noise_metric
from RCAIDE.Library.Methods.Noise.Metrics import EPNL_noise_metric
from RCAIDE.Library.Methods.Noise.Metrics import A_weighting_metric  

# Python package imports   
import numpy as np   
from copy import deepcopy

# ----------------------------------------------------------------------------------------------------------------------     
#  turbofan engine noise 
# ----------------------------------------------------------------------------------------------------------------------         
def turbofan_engine_noise(microphone_locations,turbofan,aeroacoustic_data,segment,settings):  
    """This method predicts the free-field 1/3 Octave Band SPL of coaxial subsonic
       jets for turbofan engines under the following conditions:
       a) Flyover (observer on ground)
       b) Static (observer on ground)
       c) In-flight or in-flow (observer on airplane or in a wind tunnel)

    Assumptions:
        SAE ARP876D: Gas Turbine Jet Exhaust Noise Prediction

    Inputs:
        vehicle	 - RCAIDE type vehicle 
        includes these fields:
            Velocity_primary           - Primary jet flow velocity                           [m/s]
            Temperature_primary        - Primary jet flow temperature                        [m/s]
            Pressure_primary           - Primary jet flow pressure                           [Pa]
            Area_primary               - Area of the primary nozzle                          [m^2]
            Velocity_secondary         - Secondary jet flow velocity                         [m/s]
            Temperature_secondary      - Secondary jet flow temperature                      [m/s]
            Pressure_secondary         - Secondary jet flow pressure                         [Pa]
            Area_secondary             - Area of the secondary nozzle                        [m^2]
            AOA                        - Angle of attack                                     [rad]
            Velocity_aircraft          - Aircraft velocity                                   [m/s]
            Altitude                   - Altitude                                            [m]
            N1                         - Fan rotational speed                                [rpm]
            EXA                        - Distance from fan face to fan exit/ fan diameter    [m]
            Plug_diameter              - Diameter of the engine external plug                [m]
            Engine_height              - Engine centerline height above the ground plane     [m]
            distance_microphone        - Distance from the nozzle exhaust to the microphones [m]
            angles                     - Array containing the desired polar angles           [rad] 

    Outputs: One Third Octave Band SPL [dB]
        SPL_p                           - Sound Pressure Level of the primary jet            [dB]
        SPL_s                           - Sound Pressure Level of the secondary jet          [dB]
        SPL_m                           - Sound Pressure Level of the mixed jet              [dB]
        SPL_total                       - Sound Pressure Level of the total jet noise        [dB]

    """ 
    # unpack   
    Velocity_primary       = aeroacoustic_data.core_nozzle.exit_velocity * 0.92*(turbofan.design_thrust/52700.)   
    Temperature_primary    = aeroacoustic_data.core_nozzle.exit_stagnation_temperature[:,0] 
    Pressure_primary       = aeroacoustic_data.core_nozzle.exit_stagnation_pressure[:,0]  
    Velocity_secondary     = aeroacoustic_data.fan_nozzle.exit_velocity  * (turbofan.design_thrust/52700.) 
    Temperature_secondary  = aeroacoustic_data.fan_nozzle.exit_stagnation_temperature[:,0] 
    Pressure_secondary     = aeroacoustic_data.fan_nozzle.exit_stagnation_pressure[:,0]  
    Velocity_aircraft      = segment.conditions.freestream.velocity[:,0]
    Mach_aircraft          = segment.conditions.freestream.mach_number
    Altitude               = segment.conditions.freestream.altitude[:,0] 
    AOA                    = np.mean(segment.conditions.aerodynamics.angles.alpha / Units.deg) 
    noise_time             = segment.conditions.frames.inertial.time[:,0]  
    distance_microphone    = np.linalg.norm(microphone_locations,axis = 1)
    angles                 = np.arccos(microphone_locations[:,0]/distance_microphone) 
    phi                    = np.arccos(microphone_locations[:,1]/distance_microphone)      
    
    N1                     = turbofan.fan.angular_velocity * 0.92*(turbofan.design_thrust/52700.)
    Diameter_primary       = turbofan.core_nozzle.diameter
    Diameter_secondary     = turbofan.fan_nozzle.diameter
    engine_height          = turbofan.engine_height
    EXA                    = turbofan.exa
    Plug_diameter          = turbofan.plug_diameter 
    Xe                     = turbofan.geometry_xe
    Ye                     = turbofan.geometry_ye
    Ce                     = turbofan.geometry_Ce 
    
    frequency              = settings.center_frequencies[5:]        
    n_cpts                 = len(noise_time)     
    num_f                  = len(frequency) 
    n_mic                  = len(microphone_locations)

    if type(Velocity_primary) == float:
        Velocity_primary    = np.ones(n_cpts)*Velocity_primary

    if type(Velocity_secondary) == float:
        Velocity_secondary  = np.ones(n_cpts)*Velocity_secondary

    # ==============================================
    # Computing atmospheric conditions
    # ==============================================  
    sound_ambient       = segment.conditions.freestream.speed_of_sound[:,0]
    density_ambient     = segment.conditions.freestream.density[:,0]  
    pressure_amb        = segment.conditions.freestream.pressure[:,0] 
    pressure_isa        = 101325 # [Pa]
    R_gas               = 287.1  # [J/kg K]
    gamma_primary       = 1.37  # Corretion for the primary jet
    gamma               = 1.4

    # Calculation of nozzle areas
    Area_primary   = np.pi*(Diameter_primary/2)**2 
    Area_secondary =  np.pi*(Diameter_secondary/2)**2   
   
    # Defining each array before the main loop  
    theta_P                = np.tile(np.arctan2(microphone_locations[:, 1], microphone_locations[:, 0])[None,:],(n_cpts,1))  
    theta_S                = deepcopy(theta_P)
    theta_M                = deepcopy(theta_P)
    EX_p                   = np.zeros((n_cpts,n_mic,num_f)) 
    EX_s                   = np.zeros((n_cpts,n_mic,num_f)) 
    EX_m                   = np.zeros((n_cpts,n_mic,num_f))  
    SPL_p                  = np.zeros((n_cpts,n_mic,num_f)) 
    SPL_s                  = np.zeros((n_cpts,n_mic,num_f)) 
    SPL_m                  = np.zeros((n_cpts,n_mic,num_f)) 
    SPL                    = np.zeros((n_cpts,n_mic))
    SPL_dBA                = np.zeros((n_cpts,n_mic))
    SPL_1_3_spectrum       = np.zeros((n_cpts,n_mic,num_f))
    SPL_primary_spectrum   = np.zeros((n_cpts,n_mic,num_f))
    SPL_secondary_spectrum = np.zeros((n_cpts,n_mic,num_f))
    SPL_mixed_specturm     = np.zeros((n_cpts,n_mic,num_f)) 
    SPL_1_3_spectrum_dBA   = np.zeros((n_cpts,n_mic,num_f)) 

    # Start loop for each position of aircraft 
    for i in range(n_cpts):
        for j in range(n_mic): 
        
            EX_p    = np.zeros(num_f)
            EX_s    = np.zeros(num_f)
            EX_m    = np.zeros(num_f) 
            theta_p = abs(theta_P[i,j])
            theta_s = abs(theta_S[i,j])
            theta_m = abs(theta_M[i,j])
 
            # Primary and Secondary jets
            Cpp = R_gas/(1-1/gamma_primary)
            Cp  = R_gas/(1-1/gamma)
    
            density_primary   = Pressure_primary[i]/(R_gas*Temperature_primary[i]-(0.5*R_gas*Velocity_primary[i]**2/Cpp))
            density_secondary = Pressure_secondary[i]/(R_gas*Temperature_secondary[i]-(0.5*R_gas*Velocity_secondary[i]**2/Cp))
    
            mass_flow_primary   = Area_primary*Velocity_primary[i]*density_primary
            mass_flow_secondary = Area_secondary*Velocity_secondary[i]*density_secondary
    
            #Mach number of the external flow - based on the aircraft velocity
            Mach_aircraft[i] = Velocity_aircraft[i]/sound_ambient[i]
    
            #Calculation Procedure for the Mixed Jet Flow Parameters
            Velocity_mixed = (mass_flow_primary*Velocity_primary[i]+mass_flow_secondary*Velocity_secondary[i])/ \
                (mass_flow_primary+mass_flow_secondary)
            Temperature_mixed =(mass_flow_primary*Temperature_primary[i]+mass_flow_secondary*Temperature_secondary[i])/ \
                (mass_flow_primary+mass_flow_secondary)
            density_mixed = pressure_amb[i]/(R_gas*Temperature_mixed-(0.5*R_gas*Velocity_mixed**2/Cp))
            Area_mixed = Area_primary*density_primary*Velocity_primary[i]*(1+(mass_flow_secondary/mass_flow_primary))/ \
                (density_mixed*Velocity_mixed)
            Diameter_mixed = (4*Area_mixed/np.pi)**0.5
     
    
            XBPR = mass_flow_secondary/mass_flow_primary - 5.5
            if XBPR<0:
                XBPR=0
            elif XBPR>4:
                XBPR=4
    
            #Auxiliary parameter defined as DVPS
            DVPS = np.abs((Velocity_primary[i] - (Velocity_secondary[i]*Area_secondary+Velocity_aircraft[i]*Area_primary)/\
                           (Area_secondary+Area_primary)))
            if DVPS<0.3:
                DVPS=0.3
    
            # Calculation of the Strouhal number for each jet component (p-primary, s-secondary, m-mixed)
            Str_p = frequency*Diameter_primary/(DVPS)  #Primary jet
            Str_s = frequency*Diameter_mixed/(Velocity_secondary[i]-Velocity_aircraft[i]) #Secondary jet
            Str_m = frequency*Diameter_mixed/(Velocity_mixed-Velocity_aircraft[i]) #Mixed jet
    
            #Calculation of the Excitation adjustment parameter 
            excitation_Strouhal = (N1/60)*(Diameter_mixed/Velocity_mixed)
            
            SX = 50*(excitation_Strouhal-0.25)*(excitation_Strouhal-0.5)
            SX[excitation_Strouhal > 0.25] = 0.0 
            SX[excitation_Strouhal < 0.5]  = 0.0 
    
            # Effectiveness
            exps = np.exp(-SX)
    
            #Spectral Shape Factor
            exs = 5*exps*np.exp(-(np.log10(Str_m/(2*excitation_Strouhal+0.00001)))**2)
    
            #Fan Duct Lenght Factor
            exd = np.exp(0.6-(EXA)**0.5)
    
            #Excitation source location factor (zk)
            zk = 1-0.4*(exd)*(exps)     
    
            # Loop for the frequency array range 
            exc = sound_ambient[i]/Velocity_mixed 
            if theta_m>1.4:
                exc  = (sound_ambient[i]/Velocity_mixed)*(1-(1.8/np.pi)*(theta_m-1.4))

            #Acoustic excitation adjustment (EX)
            EX_m = exd*exs*exc   # mixed component - dependant of the frequency
    
            EX_p = +5*exd*exps   #primary component - no frequency dependance
            EX_s = 2*sound_ambient[i]/(Velocity_secondary[i]*(zk)) #secondary component - no frequency dependance    
    
            distance_primary   = distance_microphone[j] 
            distance_secondary = distance_microphone[j] 
            distance_mixed     = distance_microphone[j]
    
            #Noise attenuation due to Ambient Pressure
            dspl_ambient_pressure = 20*np.log10(pressure_amb[i]/pressure_isa)
    
            #Noise attenuation due to Density Gradientes
            dspl_density_p = 20*np.log10((density_primary+density_secondary)/(2*density_ambient[i]))
            dspl_density_s = 20*np.log10((density_secondary+density_ambient[i])/(2*density_ambient[i]))
            dspl_density_m = 20*np.log10((density_mixed+density_ambient[i])/(2*density_ambient[i]))
    
            #Noise attenuation due to Spherical divergence
            dspl_spherical_p = 20*np.log10(Diameter_primary/distance_primary)
            dspl_spherical_s = 20*np.log10(Diameter_mixed/distance_secondary)
            dspl_spherical_m = 20*np.log10(Diameter_mixed/distance_mixed)
    
            # Noise attenuation due to Geometric Near-Field 
            dspl_geometric_p = 0.0
            dspl_geometric_s = 0.0
            dspl_geometric_m = 0.0 
            
    
            # Noise attenuation due to Acoustic Near-Field 
            dspl_acoustic_p = 0.0 
            dspl_acoustic_s = 0.0 
            dspl_acoustic_m = 0.0  
     
            # Atmospheric attenuation
            delta_atmo = atmospheric_attenuation(np.array([distance_primary]),frequency)[0,:]
    
            dspl_attenuation_p = -delta_atmo 
            dspl_attenuation_s = -delta_atmo 
            dspl_attenuation_m = -delta_atmo  
    
            # Calculation of the total noise attenuation (p-primary, s-secondary, m-mixed components)
            DSPL_p = dspl_ambient_pressure+dspl_density_p+dspl_geometric_p+dspl_acoustic_p+dspl_attenuation_p+dspl_spherical_p
            DSPL_s = dspl_ambient_pressure+dspl_density_s+dspl_geometric_s+dspl_acoustic_s+dspl_attenuation_s+dspl_spherical_s
            DSPL_m = dspl_ambient_pressure+dspl_density_m+dspl_geometric_m+dspl_acoustic_m+dspl_attenuation_m+dspl_spherical_m
    
    
            # Calculation of interference effects on jet noise
            ATK_m   = angle_of_attack_effect(AOA,Mach_aircraft[i],theta_m)
            INST_s  = jet_installation_effect(Xe,Ye,Ce,theta_s,Diameter_mixed)
            Plug    = external_plug_effect(Velocity_primary[i],Velocity_secondary[i], Velocity_mixed, Diameter_primary,Diameter_secondary,
                                           Diameter_mixed, Plug_diameter, sound_ambient[i], theta_p,theta_s,theta_m)
    
            GPROX_m = ground_proximity_effect(Velocity_mixed,sound_ambient[i],theta_m,engine_height,Diameter_mixed,frequency)
    
            # Calculation of the sound pressure level for each jet component
            SPL_p = primary_noise_component(Velocity_primary[i],Temperature_primary[i],R_gas,theta_p,DVPS,sound_ambient[i],
                                            Velocity_secondary[i],Velocity_aircraft[i],Area_primary,Area_secondary,DSPL_p,EX_p,Str_p) + Plug.PG_p
    
            SPL_s = secondary_noise_component(Velocity_primary[i],theta_s,sound_ambient[i],Velocity_secondary[i],
                                              Velocity_aircraft[i],Area_primary,Area_secondary,DSPL_s,EX_s,Str_s) + Plug.PG_s + INST_s
    
            SPL_m = mixed_noise_component(Velocity_primary[i],theta_m,sound_ambient[i],Velocity_secondary[i],
                                          Velocity_aircraft[i],Area_primary,Area_secondary,DSPL_m,EX_m,Str_m,Velocity_mixed,XBPR) + Plug.PG_m + ATK_m + GPROX_m
    
            # Sum of the Total Noise
            SPL_total = 10 * np.log10(10**(0.1*SPL_p)+10**(0.1*SPL_s)+10**(0.1*SPL_m))
    
            # Store SPL history      
            SPL_1_3_spectrum[i,j,:]       = SPL_total 
            SPL[i,j]                      = SPL_arithmetic(np.atleast_2d(SPL_total),sum_axis=1 )
            SPL_primary_spectrum[i,j,:]   = SPL_p 
            SPL_secondary_spectrum[i,j,:] = SPL_s 
            SPL_mixed_specturm[i,j,:]     = SPL_m  
            SPL_1_3_spectrum_dBA[i,j,:]   = A_weighting_metric(SPL_total,frequency)
            SPL_dBA[i,j]                  = SPL_arithmetic(np.atleast_2d(A_weighting_metric(SPL_total,frequency)),sum_axis=1) 

    # Calculation of the Perceived Noise Level EPNL based on the sound time history
    PNL_total               =  PNL_noise_metric(SPL_1_3_spectrum)    
    PNL_primary             =  PNL_noise_metric(SPL_primary_spectrum)  
    PNL_secondary           =  PNL_noise_metric(SPL_secondary_spectrum)  
    PNL_mixed               =  PNL_noise_metric(SPL_mixed_specturm)  

    # Calculation of the tones corrections on the SPL for each component and total
    tone_correction_total     = noise_tone_correction(SPL_1_3_spectrum) 
    tone_correction_primary   = noise_tone_correction(SPL_primary_spectrum) 
    tone_correction_secondary = noise_tone_correction(SPL_secondary_spectrum) 
    tone_correction_mixed     = noise_tone_correction(SPL_mixed_specturm) 

    # Calculation of the PLNT for each component and total
    PNLT_total     = PNL_total+tone_correction_total
    PNLT_primary   = PNL_primary+tone_correction_primary
    PNLT_secondary = PNL_secondary+tone_correction_secondary
    PNLT_mixed     = PNL_mixed+tone_correction_mixed

    # Calculation of the EPNL for each component and total
    EPNL_total     = EPNL_noise_metric(PNLT_total)
    EPNL_primary   = EPNL_noise_metric(PNLT_primary)
    EPNL_secondary = EPNL_noise_metric(PNLT_secondary)
    EPNL_mixed     = EPNL_noise_metric(PNLT_mixed) 

    # Open output file to print the result     
    
    engine_noise                   = Data() 
    engine_noise.EPNL_primary      = EPNL_primary  
    engine_noise.EPNL_secondary    = EPNL_secondary
    engine_noise.EPNL_mixed        = EPNL_mixed    
    engine_noise.EPNL_total        = EPNL_total  
    engine_noise.SPL_1_3_spectrum  = SPL_1_3_spectrum_dBA
    engine_noise.SPL               = SPL
    engine_noise.SPL_dBA           = SPL_dBA

    return engine_noise


# RCAIDE/Methods/Mission/Segments/Climb/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 

from . import Constant_Speed_Constant_Rate
from . import Constant_Mach_Constant_Rate
from . import Constant_Mach_Constant_Angle
from . import Linear_Mach_Constant_Rate
from . import Linear_Speed_Constant_Rate
from . import Constant_Throttle_Constant_Speed
from . import Constant_Dynamic_Pressure_Constant_Rate
from . import Constant_Speed_Constant_Angle
from . import Constant_Speed_Constant_Angle_Noise
from . import Constant_EAS_Constant_Rate
from . import Constant_CAS_Constant_Rate
from . import Constant_Mach_Linear_Altitude
from . import Constant_Dynamic_Pressure_Constant_Angle
from . import Constant_Speed_Linear_Altitude 


## @ingroup Input_Output-OpenVSP
# write_vsp_mesh.py
# 
# Created:  Oct 2016, T. MacDonald
# Modified: Jan 2017, T. MacDonald
#           Feb 2017, T. MacDonald
#           Jan 2019, T. MacDonald
#           Jan 2020, T. MacDonald

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
try:
    import vsp as vsp
except ImportError:
    try:
        import openvsp as vsp
    except ImportError:
        # This allows RCAIDE to build without OpenVSP
        pass
import numpy as np
import time
import fileinput

# ---------------------------------------------------------------------------------------------------------------------- 
# write_vsp_mesh
# ---------------------------------------------------------------------------------------------------------------------- 
def write_vsp_mesh(geometry,tag,half_mesh_flag,growth_ratio,growth_limiting_flag):
    """This create an .stl surface mesh based on a vehicle stored in a .vsp3 file.
    
    Assumptions:
    None

    Source:
    N/A

    Inputs:
    geometry.                                 - Also passed to set_sources
      wings.main_wing.chords.mean_aerodynamic [m]
    half_mesh_flag                            <boolean>  determines if a symmetry plane is created
    growth_ratio                              [-]        growth ratio for the mesh
    growth_limiting_flag                      <boolean>  determines if 3D growth limiting is used

    Outputs:
    <tag>.stl                               

    Properties Used:
    N/A
    """      
    
    # Reset OpenVSP to avoid including a previous vehicle
    vsp.ClearVSPModel()
    
    if 'turbofan' in geometry.networks:
        print('Warning: no meshing sources are currently implemented for the nacelle')

    # Turn on symmetry plane splitting to improve robustness of meshing process
    if half_mesh_flag == True:
        f = fileinput.input(tag + '.vsp3',inplace=1)
        for line in f:
            if 'SymmetrySplitting' in line:
                print(line[0:34] + '1' + line[35:-1])
            else:
                print(line)
    
    vsp.ReadVSPFile(tag + '.vsp3')
    
    # Set output file types and what will be meshed
    file_type = vsp.CFD_STL_TYPE + vsp.CFD_KEY_TYPE
    set_int   = vsp.SET_ALL

    vsp.SetComputationFileName(vsp.CFD_STL_TYPE, tag + '.stl')
    vsp.SetComputationFileName(vsp.CFD_KEY_TYPE, tag + '.key')
    
    # Set to create a tagged STL mesh file
    vehicle_cont = vsp.FindContainer('Vehicle',0)
    STL_multi    = vsp.FindParm(vehicle_cont, 'MultiSolid', 'STLSettings')
    vsp.SetParmVal(STL_multi, 1.0)
    
    vsp.SetCFDMeshVal(vsp.CFD_FAR_FIELD_FLAG,1)
    if half_mesh_flag == True:
        vsp.SetCFDMeshVal(vsp.CFD_HALF_MESH_FLAG,1)
        
    # Figure out the size of the bounding box
    vehicle_id = vsp.FindContainersWithName('Vehicle')[0]
    xlen = vsp.GetParmVal(vsp.FindParm(vehicle_id,"X_Len","BBox"))
    ylen = vsp.GetParmVal(vsp.FindParm(vehicle_id,"Y_Len","BBox"))
    zlen = vsp.GetParmVal(vsp.FindParm(vehicle_id,"Z_Len","BBox"))
    
    # Max length
    max_len = np.max([xlen,ylen,zlen])
    far_length = 10.*max_len
        
    vsp.SetCFDMeshVal(vsp.CFD_FAR_SIZE_ABS_FLAG,1)
    vsp.SetCFDMeshVal(vsp.CFD_FAR_LENGTH,far_length)
    vsp.SetCFDMeshVal(vsp.CFD_FAR_WIDTH,far_length)
    vsp.SetCFDMeshVal(vsp.CFD_FAR_HEIGHT,far_length)    
    vsp.SetCFDMeshVal(vsp.CFD_FAR_MAX_EDGE_LEN, max_len)
    vsp.SetCFDMeshVal(vsp.CFD_GROWTH_RATIO, growth_ratio)
    if growth_limiting_flag == True:
        vsp.SetCFDMeshVal(vsp.CFD_LIMIT_GROWTH_FLAG, 1.0)
    
    # Set the max edge length so we have on average 50 elements per chord length
    MAC     = geometry.wings.main_wing.chords.mean_aerodynamic
    min_len = MAC/50.
    vsp.SetCFDMeshVal(vsp.CFD_MAX_EDGE_LEN,min_len)
    
    # vsp.AddDefaultSources()   
    set_sources(geometry)
    
    vsp.Update()
    
    vsp.WriteVSPFile(tag + '_premesh.vsp3')
    
    print('Starting mesh for ' + tag + ' (This may take several minutes)')
    ti = time.time()
    vsp.ComputeCFDMesh(set_int,file_type)
    tf = time.time()
    dt = tf-ti
    print('VSP meshing for ' + tag + ' completed in ' + str(dt) + ' s')
    
# ----------------------------------------------------------------------------------------------------------------------     
# set_sources
# ---------------------------------------------------------------------------------------------------------------------- 
def set_sources(geometry):
    """This sets meshing sources in a way similar to the OpenVSP default. Some source values can
    also be optionally specified as below.
    
    Assumptions:
    None

    Source:
    https://github.com/OpenVSP/OpenVSP (with some modifications)

    Inputs:
    geometry.
      wings.*.                              (passed to add_segment_sources())
        tag                                 <string>
        Segments.*.percent_span_location    [-] (.1 is 10%)
        Segments.*.root_chord_percent       [-] (.1 is 10%)
        chords.root                         [m]
        chords.tip                          [m]
        vsp_mesh                            (optional) - This holds settings that are used in add_segment_sources
      fuselages.*.
        tag                                 <string>
        vsp_mesh.                           (optional)
          length                            [m]
          radius                            [m]
        lengths.total                       (only used if vsp_mesh is not defined for the fuselage)

    Outputs:
    <tag>.stl                               

    Properties Used:
    N/A
    """     
    # Extract information on geometry type (for some reason it seems VSP doesn't have a simple 
    # way to do this)
    comp_type_dict = dict()
    comp_dict      = dict()
    for wing in geometry.wings:
        comp_type_dict[wing.tag] = 'wing'
        comp_dict[wing.tag] = wing
    for fuselage in geometry.fuselages:
        comp_type_dict[fuselage.tag] = 'fuselage'
        comp_dict[fuselage.tag] = fuselage
    # network sources have not been implemented
    #for network in geometry.networks:
        #comp_type_dict[network.tag] = 'turbojet'
        #comp_dict[network.tag] = network
        
    components = vsp.FindGeoms()
    
    # The default source values are (mostly) based on the OpenVSP scripts, wing for example:
    # https://github.com/OpenVSP/OpenVSP/blob/a5ac5302b320e8e318830663bb50ba0d4f2d6f64/src/geom_core/WingGeom.cpp
    
    for comp in components:
        comp_name = vsp.GetGeomName(comp)
        if comp_name not in comp_dict:
            continue
        comp_type = comp_type_dict[comp_name]
        # Nacelle sources are not implemented
        #if comp_name[0:8] == 'turbofan':
            #comp_type = comp_type_dict[comp_name[0:8]]
        #else:
            #comp_type = comp_type_dict[comp_name]
        if comp_type == 'wing':
            wing = comp_dict[comp_name]
            if len(wing.segments) == 0: # check if segments exist
                num_secs = 1
                use_base = True
            else:
                if wing.segments[0].percent_span_location == 0.: # check if first segment starts at the root
                    num_secs = len(wing.segments)
                    use_base = False
                else:
                    num_secs = len(wing.segments) + 1
                    use_base = True
                    
            u_start = 0.
            base_root = wing.chords.root
            base_tip  = wing.chords.tip            
            for ii in range(0,num_secs):
                if (ii==0) and (use_base == True): # create sources on root segment
                    cr = base_root
                    if len(wing.segments) > 0:
                        ct = base_root  * wing.segments[0].root_chord_percent
                        seg = wing.segments[ii]
                    else:
                        if 'vsp_mesh' in wing:
                            custom_flag = True
                        else:
                            custom_flag = False
                        ct = base_tip           
                        seg = wing
                    # extract CFD source parameters
                    if len(wing.segments) == 0:
                        wingtip_flag = True
                    else:
                        wingtip_flag = False
                    add_segment_sources(comp,cr, ct, ii, u_start, num_secs, custom_flag, 
                                  wingtip_flag,seg)                        
                elif (ii==0) and (use_base == False): 
                    cr = base_root * wing.segments[0].root_chord_percent
                    if num_secs > 1:
                        ct = base_root  * wing.segments[1].root_chord_percent
                    else:
                        ct = base_tip
                    # extract CFD source parameters
                    seg = wing.segments[ii]
                    if 'vsp_mesh' in wing.segments[ii]:
                        custom_flag = True
                    else:
                        custom_flag = False
                    wingtip_flag = False
                    add_segment_sources(comp,cr, ct, ii, u_start, num_secs, custom_flag, 
                                  wingtip_flag,seg)
                elif ii < num_secs - 1:
                    if use_base == True:
                        jj = 1
                    else:
                        jj = 0
                    cr = base_root * wing.segments[ii-jj].root_chord_percent
                    ct = base_root * wing.segments[ii+1-jj].root_chord_percent
                    seg = wing.segments[ii-jj]
                    if 'vsp_mesh' in wing.segments[ii-jj]:
                        custom_flag = True
                    else:
                        custom_flag = False
                    wingtip_flag = False
                    add_segment_sources(comp,cr, ct, ii, u_start, num_secs, custom_flag, 
                                  wingtip_flag,seg)                   
                else:     
                    if use_base == True:
                        jj = 1
                    else:
                        jj = 0                    
                    cr = base_root * wing.segments[ii-jj].root_chord_percent
                    ct = base_tip
                    seg = wing.segments[ii-jj]
                    if 'vsp_mesh' in wing.segments[ii-jj]:
                        custom_flag = True
                    else:
                        custom_flag = False
                    wingtip_flag = True
                    add_segment_sources(comp,cr, ct, ii, u_start, num_secs, custom_flag, 
                                  wingtip_flag,seg)  
                pass
                    
        elif comp_type == 'fuselage':
            fuselage = comp_dict[comp_name]
            if 'vsp_mesh' in fuselage:
                len1 = fuselage.vsp_mesh.length
                rad1 = fuselage.vsp_mesh.radius
            else:
                len1 = 0.1 * 0.5 # not sure where VSP is getting this value
                rad1 = 0.2 * fuselage.lengths.total
            uloc = 0.0
            wloc = 0.0
            vsp.AddCFDSource(vsp.POINT_SOURCE,comp,0,len1,rad1,uloc,wloc) 
            uloc = 1.0
            vsp.AddCFDSource(vsp.POINT_SOURCE,comp,0,len1,rad1,uloc,wloc) 
            pass
        
        # This is a stub for the nacelle implementation. It will create sources
        # as is but they will not be appropriate for the nacelle shape.
        
        #elif comp_type == 'turbofan':
            #network = comp_dict[comp_name[0:8]]
            #if network.has_key('vsp_mesh'):
                #len1 = network.vsp_mesh.length
                #rad1 = network.vsp_mesh.radius
            #else:
                #len1 = 0.1 * 0.5 # not sure where VSP is getting this value
            #uloc = 0.0
            #wloc = 0.0
            #vsp.AddCFDSource(vsp.POINT_SOURCE,comp,0,len1,rad1,uloc,wloc) 
            #uloc = 1.0
            #vsp.AddCFDSource(vsp.POINT_SOURCE,comp,0,len1,rad1,uloc,wloc) 
            #pass        
    
            
# ----------------------------------------------------------------------------------------------------------------------         
# add_segment_sources
# ---------------------------------------------------------------------------------------------------------------------- 
def add_segment_sources(comp,cr,ct,ii,u_start,num_secs,custom_flag,wingtip_flag,seg):
    """This sets meshing sources for the wing segments according to their size and position.
    
    Assumptions:
    None

    Source:
    https://github.com/OpenVSP/OpenVSP (with some modifications)

    Inputs:
    comp             <string> - OpenVSP component ID
    cr               [m]      - root chord
    ct               [m]      - tip chord
    ii               [-]      - segment index
    u_start          [-]      - OpenVSP parameter determining the u dimensional start point
    num_secs         [-]      - number of segments on the corresponding wing
    custom_flag      <boolean> - determines if custom source settings are to be used
    wingtip_flag     <boolean> - indicates if the current segment is a wingtip
    seg.vsp_mesh.    (only used if custom_flag is True)
      inner_length   [m]       - length of inboard element edge
      outer_length   [m]       - length of outboard element edge
      inner_radius   [m]       - radius of influence for inboard source
      outer_radius   [m]       - radius of influence for outboard source

    Outputs:
    None - sources are added to OpenVSP instance                             

    Properties Used:
    N/A
    """     
    if custom_flag == True:
        len1 = seg.vsp_mesh.inner_length
        len2 = seg.vsp_mesh.outer_length
        rad1 = seg.vsp_mesh.inner_radius
        rad2 = seg.vsp_mesh.outer_radius
    else:
        len1 = 0.01 * cr
        len2 = 0.01 * ct
        rad1 = 0.2 * cr
        rad2 = 0.2 * ct
    uloc1 = ((ii+1)+u_start-1 +1)/(num_secs+2) # index additions are shown explicitly for cross-referencing with VSP code
    wloc1 = 0.5
    uloc2 = ((ii+1)+u_start +1)/(num_secs+2)
    wloc2 = 0.5
    vsp.AddCFDSource(vsp.LINE_SOURCE,comp,0,len1,rad1,uloc1,wloc1,len2,rad2,uloc2,wloc2)
    wloc1 = 0.
    wloc2 = 0.
    TE_match = True
    if (custom_flag == True) and ('matching_TE' in seg.vsp_mesh):
        if seg.vsp_mesh.matching_TE == False: # use default values if so
            vsp.AddCFDSource(vsp.LINE_SOURCE,comp,0,0.01 * cr,0.2 * cr,uloc1,wloc1,0.01 * ct,0.2 * ct,uloc2,wloc2) 
            TE_match = False
        else:
            vsp.AddCFDSource(vsp.LINE_SOURCE,comp,0,len1,rad1,uloc1,wloc1,len2,rad2,uloc2,wloc2)
    else:
        vsp.AddCFDSource(vsp.LINE_SOURCE,comp,0,len1,rad1,uloc1,wloc1,len2,rad2,uloc2,wloc2)  
    if wingtip_flag == True:
        len1 = len2
        rad1 = rad2
        wloc1 = 0.0
        wloc2 = 0.5
        uloc1 = uloc2
        if TE_match == False: # to match not custom TE if indicated
            len1 = 0.01 * ct
            rad1 = 0.2 * ct
        vsp.AddCFDSource(vsp.LINE_SOURCE,comp,0,len1,rad1,uloc1,wloc1,len2,rad2,uloc2,wloc2)    
    
if __name__ == '__main__':
    write_vsp_mesh(tag,True)



# RCAIDE/Library/Missions/Common/Update/gravity.py
# 
# 
# Created:  Jul 2023, M. Clarke


# ----------------------------------------------------------------------------------------------------------------------
#  Update Gravity
# ----------------------------------------------------------------------------------------------------------------------
def gravity(segment):
    """ Sets the gravity for each part of the mission
    
        Assumptions:
        Fixed sea level gravity, doesn't use a gravity model yet
        
        Inputs:
        segment.analyses.planet.features.sea_level_gravity [Data] 
            
        Outputs:
        state.conditions.freestream.gravity [meters/second^2]

        Properties Used:
        N/A
                                
    """      

    # unpack
    planet = segment.analyses.planet
    H      = segment.conditions.freestream.altitude
    
    # calculate
    g      = planet.compute_gravity(H)

    # pack
    segment.state.conditions.freestream.gravity[:,0] = g[:,0]

    return 


# RCAIDE/Library/Components/Systems/__init__.py 
# 

"""
Module containing aircraft system components for modeling various onboard systems 
and equipment. This module provides base system classes and specific implementations 
for avionics and other aircraft systems.
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .System   import System
from .Avionics import Avionics


# RCAIDE/Library/Components/Propulsors/Converters/Compressor.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Feb 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports   
from RCAIDE.Library.Components                      import Component  
from RCAIDE.Library.Methods.Propulsors.Converters.Supersonic_Nozzle.append_supersonic_nozzle_conditions import append_supersonic_nozzle_conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  Compressor 
# ----------------------------------------------------------------------------------------------------------------------  
class Supersonic_Nozzle(Component):
    """
    A supersonic nozzle component model for propulsion systems capable of supersonic outflow.

    Attributes
    ----------
    tag : str
        Identifier for the nozzle. Default is 'Supersonic_Nozzle'.
        
    polytropic_efficiency : float
        Efficiency of the expansion process accounting for losses. Default is 1.0.
        
    pressure_ratio : float
        Ratio of outlet to inlet pressure. Default is 1.0.
        
    pressure_recovery : float
        Ratio of recovered to ideal total pressure. Default is 1.0.
        
    max_area_ratio : float
        Maximum allowable exit to throat area ratio. Default is 1000.0.
        
    min_area_ratio : float
        Minimum allowable exit to throat area ratio. Default is 0.0.

    Notes
    -----
    The Supersonic_Nozzle class models a convergent-divergent nozzle that can 
    achieve supersonic exit flow. The model includes:
    
    * Isentropic flow relations
    * Shock wave effects
    * Variable area ratio capabilities
    * Real gas effects through efficiency terms
    * Pressure recovery modeling

    **Major Assumptions**

    * Pressure ratio and efficiency do not change with varying conditions
    * One-dimensional flow
    * Perfect gas behavior
    * Adiabatic process
    * No boundary layer separation
    * Steady flow conditions

    **Definitions**

    'Polytropic Efficiency'
        Measure of expansion process efficiency accounting for losses
    'Pressure Ratio'
        Ratio of exit static pressure to inlet total pressure
    'Pressure Recovery'
        Measure of total pressure preservation through the nozzle
    'Area Ratio'
        Ratio of exit area to throat area

    References
    ----------
    [1] Mattingly, J. D., & Boyer, K. M. (2016). Elements of propulsion: Gas 
        turbines and rockets, second edition Jack D. Mattingly, Keith M. Boyer. 
        American Institute of Aeronautics and Astronautics.

    See Also
    --------
    RCAIDE.Library.Components.Component
    RCAIDE.Library.Components.Propulsors.Converters.Expansion_Nozzle
    """
    
    def __defaults__(self):
        """ This sets the default values for the component to function.
        
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:
        None
    
        Outputs:
        None
    
        Properties Used:
        None
        """        
        
        #set the defaults
        self.tag                             = 'Supersonic_Nozzle'
        self.polytropic_efficiency           = 1.0
        self.pressure_ratio                  = 1.0
        self.pressure_recovery               = 1.0 
        self.max_area_ratio                  = 1000.
        self.min_area_ratio                  = 0.

    def append_operating_conditions(self,segment,propulsor): 
        propulsor_conditions =  segment.state.conditions.energy[propulsor.tag]
        append_supersonic_nozzle_conditions(self,segment,propulsor_conditions)
        return 


# RCAIDE/Framework/Analyses/Mission/Segments/Cruise/Constant_Dynamic_Pressure_Constant_Altitude_Loiter.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports 
from RCAIDE.Framework.Mission.Segments.Evaluate   import Evaluate 
from RCAIDE.Framework.Core                                 import Units   
from RCAIDE.Library.Mission                      import Common,Segments

# ----------------------------------------------------------------------------------------------------------------------
#  Constant_Dynamic_Pressure_Constant_Altitude_Loiter
# ----------------------------------------------------------------------------------------------------------------------  

class Constant_Dynamic_Pressure_Constant_Altitude_Loiter(Evaluate):
    """ Vehicle flies at a constant dynamic pressure at a set altitude for a fixed time. 
    
        Assumptions:
        Built off of a constant speed constant altitude segment
        
        Source:
        None
    """           
    
    def __defaults__(self):
        """ This sets the default solver flow. Anything in here can be modified after initializing a segment.
    
            Assumptions:
            None
    
            Source:
            N/A
    
            Inputs:
            None
    
            Outputs:
            None
    
            Properties Used:
            None
        """           
        
        # -------------------------------------------------------------------------------------------------------------- 
        #   User Inputs
        # -------------------------------------------------------------------------------------------------------------- 
        self.altitude          = 0.0
        self.dynamic_pressure  = None
        self.time              = 1.0 * Units.sec
        self.true_course       = 0.0 * Units.degrees 
        
        # -------------------------------------------------------------------------------------------------------------- 
        #  Mission specific processes 
        # --------------------------------------------------------------------------------------------------------------    
        initialize                         = self.process.initialize  
        initialize.conditions              = Segments.Cruise.Constant_Dynamic_Pressure_Constant_Altitude_Loiter.initialize_conditions
        iterate                            = self.process.iterate   
        iterate.unknowns.mission           = Common.Unpack_Unknowns.orientation
        iterate.unknowns.controls          = Common.Unpack_Unknowns.control_surfaces
        iterate.residuals.flight_dynamics  = Common.Residuals.flight_dynamics
        return




# RCAIDE/Library/Methods/Thermal_Management/Reservoirs/Reservoir_Tank/append_reservoir_conditions.py
# 
# Created:  Aug 2024, S. Shekar

# ---------------------------------------------------------------------------------------------------------------------- 
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
import RCAIDE
from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_reservoir_conditions
# ---------------------------------------------------------------------------------------------------------------------- 
def append_reservoir_conditions(reservoir,segment,coolant_line):
    """ Packs the initial reservoir conditions
    
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:  
             reservoir          (data structure)              [None]
           
        Outputs:
            segment
               coolant_temperature                            [Kelvin]
               
    
        Properties Used:
        None
    """        
    
    atmosphere    = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    alt           = -segment.conditions.frames.inertial.position_vector[:,2] 
    if segment.temperature_deviation != None:
        temp_dev = segment.temperature_deviation    
    atmo_data    = atmosphere.compute_values(altitude = alt,temperature_deviation=temp_dev)    

    ones_row                                                                                        = segment.state.ones_row    
    segment.state.conditions.energy[coolant_line.tag][reservoir.tag]                                 = Conditions()
    segment.state.conditions.energy[coolant_line.tag][reservoir.tag].coolant_temperature             = atmo_data.temperature[0,0]* ones_row(1)       
    return

def append_reservoir_segment_conditions(reservoir,segment,coolant_line,conditions):
    """Sets the initial reservoir conditions at the start of each segment as the last point from the previous segment 
    
        Assumptions:
        None
    
        Source:
        N/A
    
        Inputs:  
         reservoir          (data structure)              [None]
               
        Outputs:
        None
    
        Properties Used:
        None
    """    
    
    reservoir_conditions = conditions[coolant_line.tag][reservoir.tag]
    if segment.state.initials:  
        reservoir_initials                                   = segment.state.initials.conditions.energy[coolant_line.tag][reservoir.tag]
        reservoir_conditions.coolant_temperature[:,0]        = reservoir_initials.coolant_temperature[-1,0] 
    
    return


# RCAIDE/Methods/Aerodynamics/Airfoil_Panel_Method/hess_smith.py
# 
# 
# Created:  Dec 2023, M. Clarke
# Modified: Apr 2023, N. Nanjappa

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE imports   
from .panel_geometry import panel_geometry
from .infl_coeff  import infl_coeff
from .velocity_distribution import velocity_distribution

# pacakge imports  
import numpy as np  
from scipy.linalg import solve
 
# ----------------------------------------------------------------------------------------------------------------------
# hess_smith
# ---------------------------------------------------------------------------------------------------------------------- 
def hess_smith(x_coord,y_coord,alpha,Re,npanel):
    """Computes the incompressible, inviscid flow over an airfoil of  arbitrary shape using the Hess-Smith panel method.  

    Assumptions:
    None

    Source:  "An introduction to theoretical and computational        
                    aerodynamics", J. Moran, Wiley, 1984  
 
                                                     
    Inputs          
    x             -  Vector of x coordinates of the surface                  [unitess]     
    y             -  Vector of y coordinates of the surface                  [unitess]  
    alpha         -  Airfoil angle of attack                                 [radians] 
    npanel        -  Number of panels on the airfoil.  The number of nodes   [unitess] 
                      is equal to npanel+1, and the ith panel goes from node   
                      i to node i+1                                
                                                                           
    Outputs                                                      
    cl            -  Airfoil lift coefficient                         [unitless]           
    cd            -  Airfoil drag coefficient                         [unitless]      
    cm            -  Airfoil moment coefficient about the c/4         [unitless]            
    x_bar         -  Vector of x coordinates of the surface nodes     [unitless]           
    y_bar         -  Vector of y coordinates of the surface nodes     [unitless]          
    cp            -  Vector of coefficients of pressure at the nodes  [unitless]         

    Properties Used:
    N/A
    """      
    
    ncases    = len(alpha[0,:])
    ncpts     = len(Re) 
    alpha_2d  = np.repeat(alpha.T[np.newaxis,:, :], npanel, axis=0) 
    
    # generate panel geometry data for later use   
    l,st,ct,xbar,ybar,norm = panel_geometry(x_coord,y_coord,npanel,ncases,ncpts) 
    
    # compute matrix of aerodynamic influence coefficients
    ainfl         = infl_coeff(x_coord,y_coord,xbar,ybar,st,ct,npanel,ncases,ncpts) # ncases x ncpts x npanel+1 x npanel+1 
    
    # compute right hand side vector for the specified angle of attack 
    b_2d          = np.zeros((npanel+1,ncases, ncpts))
    b_2d[:-1,:,:] = st*np.cos(alpha_2d) - np.sin(alpha_2d)*ct
    b_2d[-1,:,:]  = -(ct[0,:,:]*np.cos(alpha_2d[-1,:,:]) + st[0,:,:]*np.sin(alpha_2d[-1,:,:]))-(ct[-1,:,:]*np.cos(alpha_2d[-1,:,:]) +st[-1,:,:]*np.sin(alpha_2d[-1,:,:]))
    
    qg = np.zeros((npanel+1, ncases, ncpts))  
    for i in range(ncases):
        for j in range(ncpts):
            A = ainfl[i, j, :, :]     
            B = b_2d[:, i, j]        
            

            qg[:, i, j] = solve(A, B)
    
    # compute the tangential velocity distribution at the midpoint of panels 
    vt            = velocity_distribution(qg,x_coord,y_coord,xbar,ybar,st,ct,alpha_2d,npanel,ncases,ncpts)
    
    return  xbar,ybar,vt,norm 